from nexusflowai.validators.chat_completions_input_validator import (
    ChatCompletionsInputValidator,
)
from nexusflowai.validators.completions_input_validator import CompletionsInputValidator
from nexusflowai.validators.create_with_tools import ChatCompletionToolsFlow



from typing import Optional

from openai._models import BaseModel

__all__ = ["NexusflowAICompletionUsage"]


class NexusflowAICompletionUsage(BaseModel):
    completion_tokens: int
    """Number of tokens in the generated completion."""

    prompt_tokens: int
    """Number of tokens in the prompt."""

    total_tokens: int
    """Total number of tokens used in the request (prompt + completion)."""

    latency: Optional[float] = None
    """Time taken for this request in ms. Optional to retain compatibility with OpenAI responses."""

    time_to_first_token: Optional[float] = None
    """Time until the first token was generated in ms. Only returned for streamed requests."""

    output_tokens_per_sec: Optional[float] = None
    """Output tokens per second. Only returned for streamed requests."""



from typing_extensions import Literal, Required, TypedDict

__all__ = ["NexusflowAIChatCompletionContentPartImageBytesParam"]


class NexusflowAIChatCompletionContentPartImageBytesParam(TypedDict, total=False):
    image_bytes: str
    """Base64 byte encoded image decoded as utf-8."""

    type: Required[Literal["image_bytes"]]
    """The type of the content part."""



from typing import List, Optional, Literal

from openai._models import BaseModel

from nexusflowai.types.completion_usage import NexusflowAICompletionUsage
from nexusflowai.types.chat_completion_message import NexusflowAIChatCompletionMessage


__all__ = ["NexusflowAIChatCompletion"]


class Choice(BaseModel):
    finish_reason: Literal[
        "stop", "length", "tool_calls", "content_filter", "function_call"
    ]
    """The reason the model stopped generating tokens.

    This will be `stop` if the model hit a natural stop point or a provided stop
    sequence, `length` if the maximum number of tokens specified in the request was
    reached, `content_filter` if content was omitted due to a flag from our content
    filters, `tool_calls` if the model called a tool, or `function_call`
    (deprecated) if the model called a function.
    """

    index: int
    """The index of the choice in the list of choices."""

    message: NexusflowAIChatCompletionMessage
    """A chat completion message generated by the model."""


class NexusflowAIChatCompletion(BaseModel):
    id: str
    """A unique identifier for the chat completion."""

    choices: List[Choice]
    """A list of chat completion choices.

    Can be more than one if `n` is greater than 1.
    """

    created: int
    """The Unix timestamp (in seconds) of when the chat completion was created."""

    model: str
    """The model used for the chat completion."""

    object: Literal["chat.completion"]
    """The object type, which is always `chat.completion`."""

    system_fingerprint: Optional[str] = None
    """This fingerprint represents the backend configuration that the model runs with.

    Can be used in conjunction with the `seed` request parameter to understand when
    backend changes have been made that might impact determinism.
    """

    usage: Optional[NexusflowAICompletionUsage] = None
    """Usage statistics for the completion request."""

    hints: Optional[List[str]] = None



from typing import Optional

from typing_extensions import Literal

from openai._models import BaseModel
from openai.types.completion_choice import Logprobs

from nexusflowai.types.chat_completion_message import NexusflowAIChatCompletionMessage


__all__ = ["NexusflowAICompletionChoice"]


class NexusflowAICompletionChoice(BaseModel):
    finish_reason: Literal["stop", "length", "content_filter", "tool_calls"]
    """The reason the model stopped generating tokens.

    This will be `stop` if the model hit a natural stop point or a provided stop
    sequence, `length` if the maximum number of tokens specified in the request was
    reached, `content_filter` if content was omitted due to a flag from our
    content filters, or `tool_calls` if the model hit a tool call return.
    """

    index: int

    logprobs: Optional[Logprobs]

    text: str

    message: NexusflowAIChatCompletionMessage = None
    """The message returned by the model"""

    class Config:
        """Override the base pydantic model behavior and raise exception if additional arguments are passed in the constructor"""

        extra = "forbid"



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Union, Optional
from typing_extensions import Literal

import httpx

from openai._types import NOT_GIVEN, Body, Query, Headers, NotGiven
from openai._utils import required_args, maybe_transform
from openai._resource import SyncAPIResource, AsyncAPIResource
from openai._response import to_raw_response_wrapper, async_to_raw_response_wrapper
from openai._streaming import Stream, AsyncStream
from openai._base_client import make_request_options
from openai.types.completion_create_params import CompletionCreateParams
from openai.types.chat import (
    ChatCompletionToolParam,
    ChatCompletionMessageParam,
    ChatCompletionToolChoiceOptionParam,
    completion_create_params,
)

from nexusflowai.validators import (
    CompletionsInputValidator,
    ChatCompletionsInputValidator,
    ChatCompletionToolsFlow,
)
from nexusflowai.types import (
    NexusflowAICompletion,
    NexusflowAIChatCompletion,
    NexusflowAIChatCompletionCreateParams,
)
from nexusflowai.utils import ALLOWED_MODELS, transform_client_header
from nexusflowai.resources.utils import (
    maybe_print_hints,
    maybe_parse_into_pydantic,
    batch_call,
)


if TYPE_CHECKING:
    from nexusflowai.client import NexusflowAI, AsyncNexusflowAI

__all__ = ["Completions", "AsyncCompletions"]


class Completions(SyncAPIResource):
    with_raw_response: CompletionsWithRawResponse

    def __init__(self, client: NexusflowAI) -> None:
        super().__init__(client)
        self.with_raw_response = CompletionsWithRawResponse(self)

    @required_args(["model", "prompt"], ["model", "prompt", "stream"])
    def create(
        self,
        *,
        model: Union[
            str,
            ALLOWED_MODELS,
        ],
        prompt: Union[str, List[str], List[int], List[List[int]], None],
        best_of: Optional[int] | NotGiven = NOT_GIVEN,
        echo: Optional[bool] | NotGiven = NOT_GIVEN,
        frequency_penalty: Optional[float] | NotGiven = NOT_GIVEN,
        logit_bias: Optional[Dict[str, int]] | NotGiven = NOT_GIVEN,
        logprobs: Optional[int] | NotGiven = NOT_GIVEN,
        max_tokens: Optional[int] | NotGiven = NOT_GIVEN,
        n: Optional[int] | NotGiven = NOT_GIVEN,
        presence_penalty: Optional[float] | NotGiven = NOT_GIVEN,
        seed: Optional[int] | NotGiven = NOT_GIVEN,
        stop: Union[Optional[str], List[str], None] | NotGiven = NOT_GIVEN,
        stream: Optional[Literal[False]] | Literal[True] | NotGiven = NOT_GIVEN,
        suffix: Optional[str] | NotGiven = NOT_GIVEN,
        temperature: Optional[float] | NotGiven = NOT_GIVEN,
        top_p: Optional[float] | NotGiven = NOT_GIVEN,
        user: str | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> NexusflowAICompletion | Stream[NexusflowAICompletion]:
        """
        Creates a completion for the provided prompt and parameters.

        Args:
          model: ID of the model to use. You can use the
              [List models](https://platform.openai.com/docs/api-reference/models/list) API to
              see all of your available models, or see our
              [Model overview](https://platform.openai.com/docs/models/overview) for
              descriptions of them.

          prompt: The prompt(s) to generate completions for, encoded as a string, array of
              strings, array of tokens, or array of token arrays.

              Note that <|endoftext|> is the document separator that the model sees during
              training, so if a prompt is not specified the model will generate as if from the
              beginning of a new document.

          stream: Whether to stream back partial progress. If set, tokens will be sent as
              data-only
              [server-sent events](https://developer.mozilla.org/en-US/docs/Web/API/Server-sent_events/Using_server-sent_events#Event_stream_format)
              as they become available, with the stream terminated by a `data: [DONE]`
              message.
              [Example Python code](https://cookbook.openai.com/examples/how_to_stream_completions).

          best_of: Generates `best_of` completions server-side and returns the "best" (the one with
              the highest log probability per token). Results cannot be streamed.

              When used with `n`, `best_of` controls the number of candidate completions and
              `n` specifies how many to return – `best_of` must be greater than `n`.

              **Note:** Because this parameter generates many completions, it can quickly
              consume your token quota. Use carefully and ensure that you have reasonable
              settings for `max_tokens` and `stop`.

          echo: Echo back the prompt in addition to the completion

          frequency_penalty: Number between -2.0 and 2.0. Positive values penalize new tokens based on their
              existing frequency in the text so far, decreasing the model's likelihood to
              repeat the same line verbatim.

              [See more information about frequency and presence penalties.](https://platform.openai.com/docs/guides/gpt/parameter-details)

          logit_bias: Modify the likelihood of specified tokens appearing in the completion.

              Accepts a JSON object that maps tokens (specified by their token ID in the GPT
              tokenizer) to an associated bias value from -100 to 100. You can use this
              [tokenizer tool](/tokenizer?view=bpe) (which works for both GPT-2 and GPT-3) to
              convert text to token IDs. Mathematically, the bias is added to the logits
              generated by the model prior to sampling. The exact effect will vary per model,
              but values between -1 and 1 should decrease or increase likelihood of selection;
              values like -100 or 100 should result in a ban or exclusive selection of the
              relevant token.

              As an example, you can pass `{"50256": -100}` to prevent the <|endoftext|> token
              from being generated.

          logprobs: Include the log probabilities on the `logprobs` most likely tokens, as well the
              chosen tokens. For example, if `logprobs` is 5, the API will return a list of
              the 5 most likely tokens. The API will always return the `logprob` of the
              sampled token, so there may be up to `logprobs+1` elements in the response.

              The maximum value for `logprobs` is 5.

          max_tokens: The maximum number of [tokens](/tokenizer) to generate in the completion.

              The token count of your prompt plus `max_tokens` cannot exceed the model's
              context length.
              [Example Python code](https://cookbook.openai.com/examples/how_to_count_tokens_with_tiktoken)
              for counting tokens.

          n: How many completions to generate for each prompt.

              **Note:** Because this parameter generates many completions, it can quickly
              consume your token quota. Use carefully and ensure that you have reasonable
              settings for `max_tokens` and `stop`.

          presence_penalty: Number between -2.0 and 2.0. Positive values penalize new tokens based on
              whether they appear in the text so far, increasing the model's likelihood to
              talk about new topics.

              [See more information about frequency and presence penalties.](https://platform.openai.com/docs/guides/gpt/parameter-details)

          seed: If specified, our system will make a best effort to sample deterministically,
              such that repeated requests with the same `seed` and parameters should return
              the same result.

              Determinism is not guaranteed, and you should refer to the `system_fingerprint`
              response parameter to monitor changes in the backend.

          stop: Up to 4 sequences where the API will stop generating further tokens. The
              returned text will not contain the stop sequence.

          suffix: The suffix that comes after a completion of inserted text.

          temperature: What sampling temperature to use, between 0 and 2. Higher values like 0.8 will
              make the output more random, while lower values like 0.2 will make it more
              focused and deterministic.

              We generally recommend altering this or `top_p` but not both.

          top_p: An alternative to sampling with temperature, called nucleus sampling, where the
              model considers the results of the tokens with top_p probability mass. So 0.1
              means only the tokens comprising the top 10% probability mass are considered.

              We generally recommend altering this or `temperature` but not both.

          user: A unique identifier representing your end-user, which can help OpenAI to monitor
              and detect abuse.
              [Learn more](https://platform.openai.com/docs/guides/safety-best-practices/end-user-ids).

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """
        request_body = maybe_transform(
            {
                "model": model,
                "prompt": prompt,
                "best_of": best_of,
                "echo": echo,
                "frequency_penalty": frequency_penalty,
                "logit_bias": logit_bias,
                "logprobs": logprobs,
                "max_tokens": max_tokens,
                "n": n,
                "presence_penalty": presence_penalty,
                "seed": seed,
                "stop": stop,
                "stream": stream,
                "suffix": suffix,
                "temperature": temperature,
                "top_p": top_p,
                "user": user,
            },
            completion_create_params.CompletionCreateParams,
        )
        request_body = CompletionsInputValidator(request_body=request_body).validate()

        extra_headers = transform_client_header(extra_headers=extra_headers)

        return self._post(
            "/completions",
            body=request_body,
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
            ),
            cast_to=NexusflowAICompletion,
            stream=stream or False,
            stream_cls=Stream[NexusflowAICompletion],
        )

    @required_args(["messages", "model"], ["messages", "model", "stream"])
    def create_with_tools(
        self,
        *,
        messages: List[ChatCompletionMessageParam],
        model: Union[
            str,
            ALLOWED_MODELS,
        ],
        frequency_penalty: Optional[float] | NotGiven = NOT_GIVEN,
        function_call: completion_create_params.FunctionCall | NotGiven = NOT_GIVEN,
        functions: List[completion_create_params.Function] | NotGiven = NOT_GIVEN,
        logit_bias: Optional[Dict[str, int]] | NotGiven = NOT_GIVEN,
        max_tokens: Optional[int] | NotGiven = NOT_GIVEN,
        n: Optional[int] | NotGiven = NOT_GIVEN,
        parallel_tool_calls: bool | NotGiven = NOT_GIVEN,
        presence_penalty: Optional[float] | NotGiven = NOT_GIVEN,
        response_format: completion_create_params.ResponseFormat | NotGiven = NOT_GIVEN,
        seed: Optional[int] | NotGiven = NOT_GIVEN,
        stop: Union[Optional[str], List[str]] | NotGiven = NOT_GIVEN,
        stream: Optional[Literal[False]] | Literal[True] | NotGiven = NOT_GIVEN,
        temperature: Optional[float] | NotGiven = NOT_GIVEN,
        tool_choice: ChatCompletionToolChoiceOptionParam | NotGiven = NOT_GIVEN,
        tools: List[ChatCompletionToolParam] | NotGiven = NOT_GIVEN,
        top_p: Optional[float] | NotGiven = NOT_GIVEN,
        user: str | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> NexusflowAIChatCompletion:
        """
        Convert a ChatCompletion with tools or response_format into a Completion with a raw prompt.
        """
        request_body = maybe_transform(
            {
                "messages": messages,
                "model": model,
                "frequency_penalty": frequency_penalty,
                "function_call": function_call,
                "functions": functions,
                "logit_bias": logit_bias,
                "max_tokens": max_tokens,
                "n": n,
                "parallel_tool_calls": parallel_tool_calls,
                "presence_penalty": presence_penalty,
                "response_format": response_format,
                "seed": seed,
                "stop": stop,
                "stream": stream,
                "temperature": temperature,
                "tool_choice": tool_choice,
                "tools": tools,
                "top_p": top_p,
                "user": user,
            },
            completion_create_params.CompletionCreateParams,
        )

        request_body = ChatCompletionsInputValidator(
            request_body=request_body
        ).validate()

        pythonic_request_body = {
            k: v for k, v in request_body.items() if not isinstance(v, NotGiven)
        }

        flow = ChatCompletionToolsFlow()
        request_body_to_send = flow.preprocess(pythonic_request_body)
        request_body_to_send = {
            k: NOT_GIVEN if v is None else v for k, v in request_body_to_send.items()
        }

        extra_headers = transform_client_header(extra_headers=extra_headers)

        res = self._post(
            "/completions",
            body=request_body_to_send,
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
            ),
            cast_to=NexusflowAICompletion,
            stream=stream or False,
            stream_cls=Stream[NexusflowAICompletion],
        )

        res = flow.postprocess(pythonic_request_body, res)

        maybe_print_hints(res)

        for choice in res.choices:
            parsed = choice.message.parsed
            choice.message.parsed = maybe_parse_into_pydantic(response_format, parsed)

        return res


class AsyncCompletions(AsyncAPIResource):
    with_raw_response: AsyncCompletionsWithRawResponse

    def __init__(self, client: AsyncNexusflowAI) -> None:
        super().__init__(client)
        self.with_raw_response = AsyncCompletionsWithRawResponse(self)

    @required_args(["model", "prompt"], ["model", "prompt", "stream"])
    async def create(
        self,
        *,
        model: Union[
            str,
            ALLOWED_MODELS,
        ],
        prompt: Union[str, List[str], List[int], List[List[int]], None],
        best_of: Optional[int] | NotGiven = NOT_GIVEN,
        echo: Optional[bool] | NotGiven = NOT_GIVEN,
        frequency_penalty: Optional[float] | NotGiven = NOT_GIVEN,
        logit_bias: Optional[Dict[str, int]] | NotGiven = NOT_GIVEN,
        logprobs: Optional[int] | NotGiven = NOT_GIVEN,
        max_tokens: Optional[int] | NotGiven = NOT_GIVEN,
        n: Optional[int] | NotGiven = NOT_GIVEN,
        presence_penalty: Optional[float] | NotGiven = NOT_GIVEN,
        seed: Optional[int] | NotGiven = NOT_GIVEN,
        stop: Union[Optional[str], List[str], None] | NotGiven = NOT_GIVEN,
        stream: Optional[Literal[False]] | Literal[True] | NotGiven = NOT_GIVEN,
        suffix: Optional[str] | NotGiven = NOT_GIVEN,
        temperature: Optional[float] | NotGiven = NOT_GIVEN,
        top_p: Optional[float] | NotGiven = NOT_GIVEN,
        user: str | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> NexusflowAICompletion | AsyncStream[NexusflowAICompletion]:
        """
        Creates a completion for the provided prompt and parameters.

        Args:
          model: ID of the model to use. You can use the
              [List models](https://platform.openai.com/docs/api-reference/models/list) API to
              see all of your available models, or see our
              [Model overview](https://platform.openai.com/docs/models/overview) for
              descriptions of them.

          prompt: The prompt(s) to generate completions for, encoded as a string, array of
              strings, array of tokens, or array of token arrays.

              Note that <|endoftext|> is the document separator that the model sees during
              training, so if a prompt is not specified the model will generate as if from the
              beginning of a new document.

          stream: Whether to stream back partial progress. If set, tokens will be sent as
              data-only
              [server-sent events](https://developer.mozilla.org/en-US/docs/Web/API/Server-sent_events/Using_server-sent_events#Event_stream_format)
              as they become available, with the stream terminated by a `data: [DONE]`
              message.
              [Example Python code](https://cookbook.openai.com/examples/how_to_stream_completions).

          best_of: Generates `best_of` completions server-side and returns the "best" (the one with
              the highest log probability per token). Results cannot be streamed.

              When used with `n`, `best_of` controls the number of candidate completions and
              `n` specifies how many to return – `best_of` must be greater than `n`.

              **Note:** Because this parameter generates many completions, it can quickly
              consume your token quota. Use carefully and ensure that you have reasonable
              settings for `max_tokens` and `stop`.

          echo: Echo back the prompt in addition to the completion

          frequency_penalty: Number between -2.0 and 2.0. Positive values penalize new tokens based on their
              existing frequency in the text so far, decreasing the model's likelihood to
              repeat the same line verbatim.

              [See more information about frequency and presence penalties.](https://platform.openai.com/docs/guides/gpt/parameter-details)

          logit_bias: Modify the likelihood of specified tokens appearing in the completion.

              Accepts a JSON object that maps tokens (specified by their token ID in the GPT
              tokenizer) to an associated bias value from -100 to 100. You can use this
              [tokenizer tool](/tokenizer?view=bpe) (which works for both GPT-2 and GPT-3) to
              convert text to token IDs. Mathematically, the bias is added to the logits
              generated by the model prior to sampling. The exact effect will vary per model,
              but values between -1 and 1 should decrease or increase likelihood of selection;
              values like -100 or 100 should result in a ban or exclusive selection of the
              relevant token.

              As an example, you can pass `{"50256": -100}` to prevent the <|endoftext|> token
              from being generated.

          logprobs: Include the log probabilities on the `logprobs` most likely tokens, as well the
              chosen tokens. For example, if `logprobs` is 5, the API will return a list of
              the 5 most likely tokens. The API will always return the `logprob` of the
              sampled token, so there may be up to `logprobs+1` elements in the response.

              The maximum value for `logprobs` is 5.

          max_tokens: The maximum number of [tokens](/tokenizer) to generate in the completion.

              The token count of your prompt plus `max_tokens` cannot exceed the model's
              context length.
              [Example Python code](https://cookbook.openai.com/examples/how_to_count_tokens_with_tiktoken)
              for counting tokens.

          n: How many completions to generate for each prompt.

              **Note:** Because this parameter generates many completions, it can quickly
              consume your token quota. Use carefully and ensure that you have reasonable
              settings for `max_tokens` and `stop`.

          presence_penalty: Number between -2.0 and 2.0. Positive values penalize new tokens based on
              whether they appear in the text so far, increasing the model's likelihood to
              talk about new topics.

              [See more information about frequency and presence penalties.](https://platform.openai.com/docs/guides/gpt/parameter-details)

          seed: If specified, our system will make a best effort to sample deterministically,
              such that repeated requests with the same `seed` and parameters should return
              the same result.

              Determinism is not guaranteed, and you should refer to the `system_fingerprint`
              response parameter to monitor changes in the backend.

          stop: Up to 4 sequences where the API will stop generating further tokens. The
              returned text will not contain the stop sequence.

          suffix: The suffix that comes after a completion of inserted text.

          temperature: What sampling temperature to use, between 0 and 2. Higher values like 0.8 will
              make the output more random, while lower values like 0.2 will make it more
              focused and deterministic.

              We generally recommend altering this or `top_p` but not both.

          top_p: An alternative to sampling with temperature, called nucleus sampling, where the
              model considers the results of the tokens with top_p probability mass. So 0.1
              means only the tokens comprising the top 10% probability mass are considered.

              We generally recommend altering this or `temperature` but not both.

          user: A unique identifier representing your end-user, which can help OpenAI to monitor
              and detect abuse.
              [Learn more](https://platform.openai.com/docs/guides/safety-best-practices/end-user-ids).

          extra_headers: Send extra headers

          extra_query: Add additional query parameters to the request

          extra_body: Add additional JSON properties to the request

          timeout: Override the client-level default timeout for this request, in seconds
        """

        request_body = maybe_transform(
            {
                "model": model,
                "prompt": prompt,
                "best_of": best_of,
                "echo": echo,
                "frequency_penalty": frequency_penalty,
                "logit_bias": logit_bias,
                "logprobs": logprobs,
                "max_tokens": max_tokens,
                "n": n,
                "presence_penalty": presence_penalty,
                "seed": seed,
                "stop": stop,
                "stream": stream,
                "suffix": suffix,
                "temperature": temperature,
                "top_p": top_p,
                "user": user,
            },
            completion_create_params.CompletionCreateParams,
        )

        request_body = CompletionsInputValidator(request_body=request_body).validate()

        extra_headers = transform_client_header(extra_headers=extra_headers)

        return await self._post(
            "/completions",
            body=request_body,
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
            ),
            cast_to=NexusflowAICompletion,
            stream=stream or False,
            stream_cls=AsyncStream[NexusflowAICompletion],
        )

    @required_args(["messages", "model"], ["messages", "model", "stream"])
    async def create_with_tools(
        self,
        *,
        messages: List[ChatCompletionMessageParam],
        model: Union[
            str,
            ALLOWED_MODELS,
        ],
        frequency_penalty: Optional[float] | NotGiven = NOT_GIVEN,
        function_call: completion_create_params.FunctionCall | NotGiven = NOT_GIVEN,
        functions: List[completion_create_params.Function] | NotGiven = NOT_GIVEN,
        logit_bias: Optional[Dict[str, int]] | NotGiven = NOT_GIVEN,
        max_tokens: Optional[int] | NotGiven = NOT_GIVEN,
        n: Optional[int] | NotGiven = NOT_GIVEN,
        parallel_tool_calls: bool | NotGiven = NOT_GIVEN,
        presence_penalty: Optional[float] | NotGiven = NOT_GIVEN,
        response_format: completion_create_params.ResponseFormat | NotGiven = NOT_GIVEN,
        seed: Optional[int] | NotGiven = NOT_GIVEN,
        stop: Union[Optional[str], List[str]] | NotGiven = NOT_GIVEN,
        stream: Optional[Literal[False]] | Literal[True] | NotGiven = NOT_GIVEN,
        temperature: Optional[float] | NotGiven = NOT_GIVEN,
        tool_choice: ChatCompletionToolChoiceOptionParam | NotGiven = NOT_GIVEN,
        tools: List[ChatCompletionToolParam] | NotGiven = NOT_GIVEN,
        top_p: Optional[float] | NotGiven = NOT_GIVEN,
        user: str | NotGiven = NOT_GIVEN,
        # Use the following arguments if you need to pass additional parameters to the API that aren't available via kwargs.
        # The extra values given here take precedence over values defined on the client or passed to this method.
        extra_headers: Headers | None = None,
        extra_query: Query | None = None,
        extra_body: Body | None = None,
        timeout: float | httpx.Timeout | None | NotGiven = NOT_GIVEN,
    ) -> NexusflowAIChatCompletion:  # pragma: no cover
        """
        Convert a ChatCompletion with tools or response_format into a Completion with a raw prompt.

        This function is a copy of the Sync `Completions.create_with_tools`
        """
        request_body = maybe_transform(
            {
                "messages": messages,
                "model": model,
                "frequency_penalty": frequency_penalty,
                "function_call": function_call,
                "functions": functions,
                "logit_bias": logit_bias,
                "max_tokens": max_tokens,
                "n": n,
                "parallel_tool_calls": parallel_tool_calls,
                "presence_penalty": presence_penalty,
                "response_format": response_format,
                "seed": seed,
                "stop": stop,
                "stream": stream,
                "temperature": temperature,
                "tool_choice": tool_choice,
                "tools": tools,
                "top_p": top_p,
                "user": user,
            },
            completion_create_params.CompletionCreateParams,
        )

        request_body = ChatCompletionsInputValidator(
            request_body=request_body
        ).validate()

        pythonic_request_body = {
            k: v for k, v in request_body.items() if not isinstance(v, NotGiven)
        }

        flow = ChatCompletionToolsFlow()
        request_body_to_send = flow.preprocess(pythonic_request_body)
        request_body_to_send = {
            k: NOT_GIVEN if v is None else v for k, v in request_body_to_send.items()
        }

        extra_headers = transform_client_header(extra_headers=extra_headers)

        res = await self._post(
            "/completions",
            body=request_body,
            options=make_request_options(
                extra_headers=extra_headers,
                extra_query=extra_query,
                extra_body=extra_body,
                timeout=timeout,
            ),
            cast_to=NexusflowAICompletion,
            stream=stream or False,
            stream_cls=AsyncStream[NexusflowAICompletion],
        )

        res = flow.postprocess(pythonic_request_body, res)

        maybe_print_hints(res)

        for choice in res.choices:
            parsed = choice.message.parsed
            choice.message.parsed = maybe_parse_into_pydantic(response_format, parsed)

        return res

    def batch_create(
        self,
        batch: List[CompletionCreateParams],
        num_calls_in_parallel: Optional[int] = None,
    ) -> List[Union[NexusflowAICompletion, Exception]]:
        """
        Will run the batched prompts in an async manner.
        """
        return batch_call(self.create, batch, num_calls_in_parallel)

    def batch_create_with_tools(
        self,
        batch: List[NexusflowAIChatCompletionCreateParams],
        num_calls_in_parallel: Optional[int] = None,
    ) -> List[Union[NexusflowAIChatCompletion, Exception]]:
        """
        Will run the batched prompts in an async manner.
        """
        return batch_call(self.create_with_tools, batch, num_calls_in_parallel)

    def batch_run_prompts(
        self,
        batch_prompts_str: List[str],
        params: CompletionCreateParams,
        num_calls_in_parallel: Optional[int] = None,
    ) -> List[NexusflowAICompletion]:
        """
        This function takes in a list of raw string prompts, appends other parameters (like modelname, temperature, etc) to it and creates a list of CompletionCreateParams objects. These objects are then passed to the batch_create function to be run asynchronously.

        Args:
        batch_prompts_str: A list of raw prompt strings
        params: A CompletionCreateParams object which is essentially a TypedDict. The key - value pairs should correspond to the input parameters - parameter types of the AsynCompletions.create function.

        Returns:
        A list of NexusflowAICompletion objects.
        """

        if params.get("stream", False):
            raise ValueError("Batch create does not support streaming requests!")

        batched_prompts: List[CompletionCreateParams] = [
            {**params, "prompt": prompt_str} for prompt_str in batch_prompts_str
        ]

        return self.batch_create(batched_prompts, num_calls_in_parallel)


class CompletionsWithRawResponse:
    def __init__(self, completions: Completions) -> None:
        self.create = to_raw_response_wrapper(
            completions.create,
        )


class AsyncCompletionsWithRawResponse:
    def __init__(self, completions: AsyncCompletions) -> None:
        self.create = async_to_raw_response_wrapper(
            completions.create,
        )



from typing import List, Literal, Optional, Type, Union

from dataclasses import dataclass

import json

from jsonschema.exceptions import SchemaError
from jsonschema.validators import validator_for

from pydantic import BaseModel, TypeAdapter

from openai.types import CompletionCreateParams
from openai.types.chat.completion_create_params import (
    CompletionCreateParams as ChatCompletionCreateParams,
)

from nexusflowai.types import (
    NexusflowAICompletion,
    NexusflowAIChatCompletion,
    NexusflowAIChatCompletionMessage,
)
from nexusflowai.types.chat_completion import Choice as NexusflowAIChatCompletionChoice

from nexusflowai.validators.create_with_tools.preprocess import FunctionCallTranslator
from nexusflowai.validators.create_with_tools.postprocess import (
    FunctionCallResponseTranslator,
    ResponseFormatTranslator,
)


class CompletionCreateParamsNonStreaming(BaseModel):
    prompt: str
    model: str
    frequency_penalty: Optional[Literal[None]] = None
    logprobs: Optional[Literal[None]] = None
    max_tokens: Optional[int] = None
    n: Optional[Literal[1]] = None
    presence_penalty: Optional[Literal[None]] = None
    seed: Optional[Literal[None]] = None
    stop: Optional[Union[str, List[str]]] = None
    temperature: Optional[float] = None
    top_p: Optional[Union[Literal[None], Literal[1.0]]] = None
    user: str = ""
    best_of: Optional[Literal[1]] = None
    echo: Optional[Literal[False]] = None
    logit_bias: Optional[Literal[None]] = None
    suffix: Optional[Literal[None]] = None
    stream: Optional[Union[Literal[False], Literal[None]]] = False


@dataclass
class ChatCompletionToolsFlow:
    MAX_NUM_FUNCTIONS_FOR_HINT = 15

    def _validate_to_completions(
        self, create_params: ChatCompletionCreateParams, target_cls: Type[BaseModel]
    ) -> CompletionCreateParams:
        create_params = target_cls.model_validate(create_params)
        create_params = create_params.model_dump()

        ta = TypeAdapter(CompletionCreateParams)
        create_params = ta.validate_python(create_params)

        return create_params

    def preprocess(
        self, create_params: ChatCompletionCreateParams
    ) -> CompletionCreateParams:
        tools = create_params.get("tools")
        response_format = create_params.get("response_format")
        assert (
            tools or response_format
        ), "Expected at least one of `tools` or `response_format`."

        if tools:
            for tool in tools:
                function_parameters = tool["function"]["parameters"]
                schema_validator = validator_for(function_parameters)
                try:
                    schema_validator.check_schema(function_parameters)
                except SchemaError as exception:
                    raise TypeError(exception.message) from exception

        fct = FunctionCallTranslator()
        prompt = fct.create_params_to_fc_prompt(create_params)
        create_params["prompt"] = prompt

        BOT_END = "<bot_end>"
        current_stop = create_params.get("stop")
        if isinstance(current_stop, str):
            create_params["stop"] = list(set([current_stop, BOT_END]))
        elif isinstance(current_stop, list):
            create_params["stop"] = list(set(current_stop + [BOT_END]))
        else:
            create_params["stop"] = [BOT_END]

        current_temperature = create_params.get("temperature")
        if current_temperature is None:
            create_params["temperature"] = 0.0

        return self._validate_to_completions(
            create_params, CompletionCreateParamsNonStreaming
        )

    def postprocess(
        self, create_params: ChatCompletionCreateParams, response: NexusflowAICompletion
    ) -> NexusflowAIChatCompletion:
        raw_response = response.choices[0].text

        tools = create_params.get("tools")
        response_format = create_params.get("response_format")
        tool_choice = create_params.get("tool_choice")

        hints = []
        tool_calls = None
        parsed = None
        nexusflowai_extras = None
        content = None
        if response_format and "extract_item" in raw_response:
            rft = ResponseFormatTranslator()
            parsed = rft.raw_response_to_parsed(response_format, raw_response)
            content = json.dumps(parsed)
            nexusflowai_extras = json.dumps({"original_plan": raw_response})
        elif tools:
            fcrt = FunctionCallResponseTranslator()
            tool_calls, processed_response = fcrt.raw_response_to_tool_calls(
                tools, raw_response, tool_choice
            )
            tool_calls = tool_calls or None
            if tool_calls is not None:
                nexusflowai_extras = json.dumps({"original_plan": processed_response})

            if len(tools) > self.MAX_NUM_FUNCTIONS_FOR_HINT:
                num_functions_hint = f"Hint: {response.model} performance is optimal within {self.MAX_NUM_FUNCTIONS_FOR_HINT} input functions!"
                hints.append(num_functions_hint)

        hints = hints or None

        return NexusflowAIChatCompletion(
            id=response.id,
            choices=[
                NexusflowAIChatCompletionChoice(
                    finish_reason="tool_calls",
                    index=0,
                    message=NexusflowAIChatCompletionMessage(
                        content=content,
                        refusal=None,
                        role="assistant",
                        tool_calls=tool_calls,
                        parsed=parsed,
                        nexusflowai_extras=nexusflowai_extras,
                    ),
                )
            ],
            created=response.created,
            model=response.model,
            object="chat.completion",
            system_fingerprint=None,
            usage=response.usage,
            hints=hints,
        )



from typing_extensions import Literal, Optional, Union

from openai._models import BaseModel

from openai.types.chat import (
    ChatCompletionContentPartTextParam,
    ChatCompletionContentPartImageParam,
)

from nexusflowai.types.chat_completion_content_part_image_bytes_param import (
    NexusflowAIChatCompletionContentPartImageBytesParam,
)

__all__ = ["NexusflowAIChatCompletionMessageToolCall", "Function"]


class Function(BaseModel):
    arguments: str
    """
    The arguments to call the function with, as generated by the model in JSON
    format. Note that the model does not always generate valid JSON, and may
    hallucinate parameters not defined by your function schema. Validate the
    arguments in your code before calling your function.
    """

    name: str
    """The name of the function to call."""


class NexusflowAIChatCompletionMessageToolCall(BaseModel):
    id: str
    """The ID of the tool call."""

    function: Function
    """The function that the model called."""

    type: Literal["function"]
    """The type of the tool. Currently, only `function` is supported."""

    execution_result: Optional[
        Union[
            ChatCompletionContentPartTextParam,
            ChatCompletionContentPartImageParam,
            NexusflowAIChatCompletionContentPartImageBytesParam,
        ]
    ] = None
    """The execution_result of the tool call. Optional."""



from __future__ import annotations

from typing import TYPE_CHECKING

from openai._resource import SyncAPIResource, AsyncAPIResource

from nexusflowai.resources.chat.completions import (
    Completions,
    AsyncCompletions,
    CompletionsWithRawResponse,
    AsyncCompletionsWithRawResponse,
)

if TYPE_CHECKING:
    from nexusflowai.client import NexusflowAI, AsyncNexusflowAI


__all__ = ["Chat", "AsyncChat"]


class Chat(SyncAPIResource):
    completions: Completions
    with_raw_response: ChatWithRawResponse

    def __init__(self, client: NexusflowAI) -> None:
        super().__init__(client)
        self.completions = Completions(client)
        self.with_raw_response = ChatWithRawResponse(self)


class AsyncChat(AsyncAPIResource):
    completions: AsyncCompletions
    with_raw_response: AsyncChatWithRawResponse

    def __init__(self, client: AsyncNexusflowAI) -> None:
        super().__init__(client)
        self.completions = AsyncCompletions(client)
        self.with_raw_response = AsyncChatWithRawResponse(self)


class ChatWithRawResponse:
    def __init__(self, chat: Chat) -> None:
        self.completions = CompletionsWithRawResponse(chat.completions)


class AsyncChatWithRawResponse:
    def __init__(self, chat: AsyncChat) -> None:
        self.completions = AsyncCompletionsWithRawResponse(chat.completions)



from dataclasses import dataclass

from typing import Optional, Type, Any, Union

from openai._types import NOT_GIVEN
from openai.types.chat.completion_create_params import (
    CompletionCreateParams as ChatCompletionCreateParams,
)
from openai.types.completion_create_params import CompletionCreateParams

BOT_END = "<bot_end>"
MAX_STOP_SEQUENCES = 4


@dataclass
class BaseInputValidator:
    request_body: Union[ChatCompletionCreateParams, CompletionCreateParams]

    def validate(self):
        raise NotImplementedError

    @classmethod
    def _validate_stop_token(
        cls, params: Union[ChatCompletionCreateParams, CompletionCreateParams]
    ):
        stop = params["stop"]
        if stop is NOT_GIVEN:
            stop = [BOT_END]
        elif isinstance(stop, str):
            stop = [stop, BOT_END] if stop != BOT_END else [BOT_END]
        elif isinstance(stop, list):
            if BOT_END not in stop:
                if len(stop) < MAX_STOP_SEQUENCES:
                    stop.append(BOT_END)
                else:
                    raise ValueError(
                        f"A maximum of {MAX_STOP_SEQUENCES} stop sequences can be used, and one of them should be {BOT_END}"
                    )
            elif len(stop) >= MAX_STOP_SEQUENCES:
                raise ValueError(
                    f"A maximum of {MAX_STOP_SEQUENCES} stop sequences can be used"
                )
        else:
            raise TypeError("Expected stop to be of type 'str' or 'list'")

        params["stop"] = stop

    @classmethod
    def _validate_stream(  # pylint: disable=unused-argument
        cls, params: Union[ChatCompletionCreateParams, CompletionCreateParams]
    ):
        return

    @classmethod
    def _validate_temperature(
        cls, params: Union[ChatCompletionCreateParams, CompletionCreateParams]
    ):
        if params.get("temperature") is NOT_GIVEN:
            params["temperature"] = 0.0

    @classmethod
    def _check_for_unsupported_parameter_value(
        cls,
        params: Union[ChatCompletionCreateParams, CompletionCreateParams],
        parameter_name: str,
        exception_class: Type[BaseException],
        supported_value: Optional[Any] = None,
        assign_supported_value: bool = False,
        exception_message: Optional[str] = None,
    ):
        """Raise an exception if the value of a parameter is not supported."""

        parameter_value = params.get(parameter_name)
        if parameter_value is NOT_GIVEN:
            if assign_supported_value:
                params[parameter_name] = supported_value
            return

        supported_value_present = supported_value is not None
        if supported_value_present and parameter_value == supported_value:
            return

        if exception_message is not None:
            message = exception_message
        elif supported_value_present:
            message = (
                f"The only supported value for the {parameter_name} parameter is "
                f"{supported_value}"
            )
        else:
            message = f"The {parameter_name} parameter is not supported"

        raise exception_class(message)



from typing import List, Literal, TypeVar, Optional

from openai._models import BaseModel

from nexusflowai.types.chat_completion_message_tool_call import (
    NexusflowAIChatCompletionMessageToolCall,
)


ContentType = TypeVar("ContentType")


class NexusflowAIChatCompletionMessage(BaseModel):
    """
    NexusflowAI chat completion message. Currently the same as openai ChatCompletionMessage
    """

    content: Optional[str]
    """The contents of the message."""

    refusal: Optional[str] = None
    """The refusal message generated by the model."""

    role: Literal["assistant"]
    """The role of the author of this message."""

    tool_calls: Optional[List[NexusflowAIChatCompletionMessageToolCall]] = None
    """The tool calls generated by the model, such as function calls."""

    parsed: Optional[ContentType] = None
    """The auto-parsed message contents"""

    nexusflowai_extras: Optional[str] = None
    """Extra items passed in for Nexusflow client, formatted as a json string."""



from typing import Union

from dataclasses import dataclass

import json

from openai._types import NOT_GIVEN
from openai.types.chat.completion_create_params import (
    CompletionCreateParams as ChatCompletionCreateParams,
)
from openai.types.completion_create_params import CompletionCreateParams

from nexusflowai.validators.base_input_validator import BaseInputValidator
from nexusflowai._exceptions import DeprecationError
from nexusflowai.validators.json_schema_to_dataclasses import (
    try_convert_to_dataclasses_str,
)


@dataclass
class ChatCompletionsInputValidator(BaseInputValidator):
    def validate(self) -> ChatCompletionCreateParams:
        tools = self.request_body["tools"]
        has_tools = tools is not NOT_GIVEN and bool(tools)
        has_response_format = self.request_body["response_format"] is not NOT_GIVEN

        if has_tools and has_response_format:
            raise NotImplementedError(
                "Using `tools` and `response_format` is currently not supported!"
            )

        self._validate_stream(params=self.request_body)

        self._validate_stop_token(params=self.request_body)

        self._check_for_unsupported_parameter_value(
            params=self.request_body,
            parameter_name="function_call",
            supported_value=NOT_GIVEN,
            exception_class=DeprecationError,
            exception_message="The function_call parameter is deprecated",
        )
        self._check_for_unsupported_parameter_value(
            params=self.request_body,
            parameter_name="functions",
            supported_value=NOT_GIVEN,
            exception_class=DeprecationError,
            exception_message="The functions parameter is deprecated in favor of tools",
        )

        self._check_for_unsupported_parameter_value(
            params=self.request_body,
            parameter_name="n",
            exception_class=NotImplementedError,
            supported_value=1,
            assign_supported_value=True,
        )

        self._validate_temperature(params=self.request_body)

        self._validate_response_format(params=self.request_body)

        return self.request_body

    @classmethod
    def _validate_stream(
        cls, params: Union[ChatCompletionCreateParams, CompletionCreateParams]
    ):
        if not params["stream"]:
            return

        if params.get("tools", []):
            raise AssertionError("Streaming is not supported with tools!")
        if params.get("response_format"):
            raise AssertionError("Streaming is not supported with response format!")

    @classmethod
    def _validate_response_format(cls, params: ChatCompletionCreateParams) -> None:
        response_format = params.get("response_format")
        if response_format is NOT_GIVEN:
            return

        json_schema_str, dataclasses_str, _ = try_convert_to_dataclasses_str(
            response_format
        )
        assert (
            dataclasses_str
        ), "The `response_format` parameter currently only supports subclasses of Pydantic BaseModel or input JSON schema!"

        json_schema = {
            "type": "json_schema",
            "json_schema": json.loads(json_schema_str),
        }
        params["response_format"] = json_schema



from typing import Any, Dict, List, Tuple, Optional, Union

from dataclasses import dataclass

from copy import deepcopy

from keyword import iskeyword

from ast import AST, parse as ast_parse
from ast import Call, Import, ImportFrom, NodeVisitor, Name

import json
import json_repair

from openai.types.chat import (
    ChatCompletionToolParam,
    ChatCompletionMessageToolCallParam,
    CompletionCreateParams,
    ChatCompletionMessageParam,
    ChatCompletionContentPartParam,
    ChatCompletionContentPartImageParam,
    ChatCompletionContentPartInputAudioParam,
    ChatCompletionContentPartRefusalParam,
)
from openai.types.chat.chat_completion_tool_choice_option_param import (
    ChatCompletionNamedToolChoiceParam,
)


ALLOWED_ROLES = [
    "system",
    "tool",
    "user",
    "assistant",
]

ALLOWED_IMPORTS = [
    "typing",
    "itertools",
    "collections",
    "tabulate",
    "dataclasses",
    "requests",
]
DISALLOWED_FUNCTIONS = ["eval", "exec", "setattr", "locals", "globals"]


class SecurityVisitor(NodeVisitor):
    def visit_Call(self, node: Call):
        if isinstance(node.func, Name):
            func_name = node.func.id
            if func_name in DISALLOWED_FUNCTIONS:
                raise ValueError(
                    f"Found dangerous call to {func_name} at line {node.lineno}. Disallowed functions: {DISALLOWED_FUNCTIONS}"
                )

        self.generic_visit(node)

    def visit_Import(self, node: Import) -> Any:
        import_names = [a.name for a in node.names]
        self.check_imports(node, import_names)

        self.generic_visit(node)

    def visit_ImportFrom(self, node: ImportFrom) -> Any:
        import_names = [node.module]
        self.check_imports(node, import_names)

        self.generic_visit(node)

    def check_imports(
        self, node: Union[Import, ImportFrom], import_names: List[str]
    ) -> None:
        disallowed_import_names = [n for n in import_names if n not in ALLOWED_IMPORTS]
        if disallowed_import_names:
            raise ImportError(
                f"Not allowed to import {disallowed_import_names} at line {node.lineno}. Allowed imports include {ALLOWED_IMPORTS}"
            )


class CodeExecutionHelper:
    def clean_input(self, source: str) -> Tuple[str, Optional[AST]]:
        source = source.strip()
        if len(source) == 0:
            return source, None

        try:
            tree = ast_parse(source)
        except SyntaxError as e:
            new_source = self._try_fix_syntax_error(e, source)
            try:
                tree = ast_parse(new_source)
                source = new_source
            except:  # pylint: disable=bare-except
                tree = None
        except:  # pylint: disable=bare-except
            tree = None

        return source, tree

    @staticmethod
    def _try_fix_syntax_error(error: SyntaxError, source: str) -> str | None:
        res = None
        if not error.args:
            return res

        error_message = error.args[0]
        res = source
        if error_message == "'(' was never closed":
            res = f"{source})"
        elif error_message == "unmatched ')'":
            res = source.removesuffix(")")
        else:
            """
            Unescaped quotes can manifest as a variety of error messages including:
            - unterminated string literal
            - Perhaps you forgot a comma?

            Hence we use this as a fallback case
            """

            possible_first_quotes = {'="': '")', "='": "')"}
            for possible_first_quote, last_quote in possible_first_quotes.items():
                if possible_first_quote not in source:
                    continue

                res = source.split(possible_first_quote, maxsplit=1)
                res = "='''".join(res)
                res = res.split(last_quote, maxsplit=1)
                res = "''')".join(res)

        return res

    def check_security(self, tree: AST) -> None:
        security_visitor = SecurityVisitor()
        try:
            security_visitor.visit(tree)
        except ValueError as e:
            raise TypeError from e


"""
A smattering of symbols that are invalid for function and argument names
These are all replaced with "_" downstream.
This is a translation table used in `str.translate`.
"""
INVALID_NAME_CHARS = str.maketrans(dict.fromkeys(" !%&(*+,-/<=>@[^|~", "_"))


def clean_str(s: str) -> str:
    s = s.translate(INVALID_NAME_CHARS)
    if iskeyword(s):
        s = f"_{s}"
    return s


def clean_raw_response(raw_response: str) -> str:
    response = raw_response.split("Thought:")[0]
    response = response.strip()
    return response


@dataclass
class CleanedFunction:
    dirty_name: str
    dirty_argument_names: List[str]

    def __post_init__(self) -> None:
        self.clean_name = clean_str(self.dirty_name)
        self.clean_argument_names = list(map(clean_str, self.dirty_argument_names))

        self.argument_names_d2c = dict(
            zip(self.dirty_argument_names, self.clean_argument_names)
        )
        self.argument_names_c2d = dict(
            zip(self.clean_argument_names, self.dirty_argument_names)
        )

    def clean_tool_param(
        self, tool: ChatCompletionToolParam
    ) -> ChatCompletionToolParam:
        tool = deepcopy(tool)

        function = tool["function"]
        function["name"] = self.clean_name

        properties = function["parameters"].get("properties")
        if properties:
            properties = {
                self.argument_names_d2c.get(k, k): v for k, v in properties.items()
            }
            function["parameters"]["properties"] = properties

        required = function["parameters"].get("required")
        if required:
            function["parameters"]["required"] = [
                self.argument_names_d2c.get(k, k) for k in required
            ]

        return tool

    def clean_tool_call_param(
        self, tool_call: ChatCompletionMessageToolCallParam
    ) -> ChatCompletionMessageToolCallParam:
        tool_call = deepcopy(tool_call)

        function = tool_call["function"]
        function["name"] = self.clean_name

        arguments = json_repair.loads(function["arguments"])
        arguments = {self.argument_names_d2c.get(k, k): v for k, v in arguments.items()}
        function["arguments"] = json.dumps(arguments)

        return tool_call

    def dirty_function_call_dict(
        self, clean_name: str, clean_kwargs: Dict[str, Any]
    ) -> Tuple[str, Dict[str, Any]]:
        assert self.clean_name == clean_name
        return self.dirty_name, {
            self.argument_names_c2d.get(k, k): v for k, v in clean_kwargs.items()
        }

    @classmethod
    def from_tool_param(cls, tool: ChatCompletionToolParam) -> "CleanedFunction":
        function = tool["function"]
        function_properties = function["parameters"].get("properties", dict())
        cf = cls(
            dirty_name=function["name"],
            dirty_argument_names=list(function_properties),
        )

        return cf


def get_tool_choice_prefix(
    tool_choice: ChatCompletionNamedToolChoiceParam | None,
    dirty_function_name_to_cf: Dict[str, CleanedFunction],
) -> str | None:
    if not isinstance(tool_choice, dict):
        return None

    tool_choice: ChatCompletionNamedToolChoiceParam
    dirty_function_name = tool_choice["function"]["name"]

    assert (
        dirty_function_name in dirty_function_name_to_cf
    ), f"Did not find any tools with the `tool_choice` function name `{dirty_function_name}`"

    cleaned_name = dirty_function_name_to_cf[dirty_function_name].clean_name

    return cleaned_name


def _parse_content_parts(
    content_parts: List[
        Union[ChatCompletionContentPartParam, ChatCompletionContentPartRefusalParam]
    ]
) -> List[str]:
    if content_parts == []:
        return []

    content_strs = []

    for part in content_parts:
        part_type = part["type"]

        match part_type:
            case "text":
                text = part["text"]
                content_strs.append(text)

            case "image_url":
                raise NotImplementedError(
                    f"{ChatCompletionContentPartImageParam} is not supported!"
                )

            case "input_audio":
                raise NotImplementedError(
                    f"{ChatCompletionContentPartInputAudioParam} is not supported!"
                )

            case "refusal":
                raise NotImplementedError(
                    f"{ChatCompletionContentPartRefusalParam} ist not supported!"
                )

    return content_strs


def _flatten_chat_completion_create_params_messages_content_parts(
    create_params: CompletionCreateParams,
) -> CompletionCreateParams:
    """
    Expands ContentPartParams to individual messages.
    Used for chat flows in backend.
    """

    def _parse_chat_message_content_parts(
        message: ChatCompletionMessageParam,
        content_parts: List[
            Union[ChatCompletionContentPartParam, ChatCompletionContentPartRefusalParam]
        ],
    ) -> List[ChatCompletionMessageParam]:
        # Expanded messages:
        # The last message is original message iwth all calls / inputs.
        # The other messages only have the role and content information.
        messages: List[ChatCompletionMessageParam] = [
            *[
                {
                    "content": None,
                    "role": message["role"],
                }
                for _ in range(len(content_parts) - 1)
            ],
            message,
        ]
        if "name" in message:
            for sub_message in messages:
                sub_message["name"] = message["name"]

        for i, content_str in enumerate(_parse_content_parts(content_parts)):
            messages[i]["content"] = content_str

        return messages

    def _expand_chat_message_content(
        message: ChatCompletionMessageParam,
    ) -> List[ChatCompletionMessageParam]:
        content = message.get("content")

        if isinstance(content, str):
            return [message]

        if content is None:
            content = []

        result = _parse_chat_message_content_parts(message, content)

        return result

    def _expand_chat_messages(
        messages: List[ChatCompletionMessageParam],
    ) -> List[ChatCompletionMessageParam]:
        conversation = []

        for message in messages:
            sub_messages = _expand_chat_message_content(message)
            conversation.extend(sub_messages)

        return conversation

    create_params["messages"] = _expand_chat_messages(create_params["messages"])

    return create_params
